@@ -212,6 +212,9 @@ ext2fs_copy_inode(EXT2FS_INFO *ext2fs, FS_INODE *fs_inode)
     int     i;
 	ext2fs_inode 	*in = ext2fs->dinode;
 	FS_INFO *fs = (FS_INFO *)&ext2fs->fs_info;
+    GRPNUM_T grpnum;
+    UCHAR  *imap = 0;
+    INUM_T  ibase = 0;
 
 	fs_inode->mode =  getu16(fs, in->i_mode);
 	fs_inode->nlink =  getu16(fs, in->i_nlink);
@@ -223,6 +226,7 @@ ext2fs_copy_inode(EXT2FS_INFO *ext2fs, FS_INODE *fs_inode)
 	fs_inode->ctime =  getu32(fs, in->i_ctime);
 	fs_inode->dtime =  getu32(fs, in->i_dtime);
 
+	fs_inode->seq = 0;
 
 	if (fs_inode->link) {
 		free(fs_inode->link);
@@ -249,7 +253,7 @@ ext2fs_copy_inode(EXT2FS_INFO *ext2fs, FS_INODE *fs_inode)
 	  (fs_inode->size < EXT2FS_MAXPATHLEN) && (fs_inode->size >= 0)) {
         int count = 0, j;
 
-        fs_inode->link = mymalloc(fs_inode->size+1);
+        fs_inode->link = mymalloc(fs_inode->size + 1);
 
         /* it is located directly in the pointers */
         if (fs_inode->size < 4 * (EXT2FS_NDADDR + EXT2FS_NIADDR)) {
@@ -283,32 +287,20 @@ ext2fs_copy_inode(EXT2FS_INFO *ext2fs, FS_INODE *fs_inode)
                 fs_read_block(fs, fs_buf, fs->block_size,
                   fs_inode->direct_addr[i], "link block");
 
-                strncpy (ptr, fs_buf->data, read_count);
+                memcpy (ptr, fs_buf->data, read_count);
                 count += read_count;
                 ptr = (char *)(int)ptr + count;
              }
+
+			 /* terminate the string */
+			 *ptr = '\0';
              fs_buf_free(fs_buf);
         }
 
     }
-}
-
-/* ext2fs_inode_lookup - lookup inode, external interface */
-
-static FS_INODE *
-ext2fs_inode_lookup(FS_INFO *fs, INUM_T inum)
-{
-    GRPNUM_T grpnum;
-    UCHAR  *imap = 0;
-    INUM_T  ibase = 0;
-    EXT2FS_INFO *ext2fs = (EXT2FS_INFO *) fs;
-    FS_INODE *fs_inode = fs_inode_alloc(EXT2FS_NDADDR,EXT2FS_NIADDR);
-
-    ext2fs_dinode_lookup(ext2fs, inum);
-    ext2fs_copy_inode(ext2fs, fs_inode);
 
 	/* Fill in the flags value */
-    grpnum = (inum - fs->first_inum) / 
+    grpnum = (ext2fs->inum - fs->first_inum) / 
 	  getu32(fs, ext2fs->fs->s_inodes_per_group);
 
 	imap = ext2fs_imap_lookup(ext2fs, grpnum);
@@ -318,7 +310,7 @@ ext2fs_inode_lookup(FS_INFO *fs, INUM_T inum)
     /*
      * Apply the allocated/unallocated restriction.
      */
-    fs_inode->flags = (isset(imap, inum - ibase) ?
+    fs_inode->flags = (isset(imap, ext2fs->inum - ibase) ?
            FS_FLAG_ALLOC : FS_FLAG_UNALLOC);
 
     fs_inode->flags |= (fs_inode->nlink ?
@@ -331,6 +323,19 @@ ext2fs_inode_lookup(FS_INFO *fs, INUM_T inum)
     fs_inode->flags |= (fs_inode->ctime ?
       FS_FLAG_USED : FS_FLAG_UNUSED);
 
+}
+
+/* ext2fs_inode_lookup - lookup inode, external interface */
+
+static FS_INODE *
+ext2fs_inode_lookup(FS_INFO *fs, INUM_T inum)
+{
+    EXT2FS_INFO *ext2fs = (EXT2FS_INFO *) fs;
+    FS_INODE *fs_inode = fs_inode_alloc(EXT2FS_NDADDR,EXT2FS_NIADDR);
+
+    ext2fs_dinode_lookup(ext2fs, inum);
+    ext2fs_copy_inode(ext2fs, fs_inode);
+
 
     return (fs_inode);
 }
@@ -961,6 +966,147 @@ ext2fs_fsstat(FS_INFO *fs, FILE *hFile)
 }
 
 
+/************************* istat *******************************/
+ 
+static int printidx = 0;
+#define WIDTH   8
+  
+/* indirect block accounting */
+#define INDIR_SIZ   64  
+static DADDR_T indirl[INDIR_SIZ];
+static  unsigned char   indir_idx;
+
+static u_int8_t
+print_addr_act (FS_INFO *fs, DADDR_T addr, char *buf,
+  int size, int flags, char *ptr)
+{  
+    FILE *hFile = (FILE *)ptr;
+
+    if (flags & FS_FLAG_CONT) {
+        int i, s;
+        /* cycle through the fragments if they exist */
+        for (i = 0, s = size; s > 0; s-= fs->block_size, i++) {
+
+            /* sparse file */
+            if (addr)
+                fprintf(hFile, "%lu ", (unsigned long) addr + i);
+            else 
+                fprintf (hFile, "0 ");
+
+            if (++printidx == WIDTH) {
+                fprintf(hFile, "\n");
+                printidx = 0;
+            }
+        }
+    }
+
+    /* this must be an indirect block pointer, so put it in the list */
+    else if (flags & FS_FLAG_META) {
+        if (indir_idx < INDIR_SIZ)
+            indirl[indir_idx++] = addr;
+    }
+    return WALK_CONT;
+}
+
+
+static void
+ext2fs_istat (FS_INFO *fs, FILE *hFile, INUM_T inum, int numblock)
+{
+    EXT2FS_INFO *ext2fs = (EXT2FS_INFO *) fs;
+    FS_INODE *fs_inode;
+    char ls[12];
+    int flags = (FS_FLAG_AONLY | FS_FLAG_CONT | FS_FLAG_META |
+          FS_FLAG_ALLOC | FS_FLAG_UNALLOC );
+
+    fs_inode = ext2fs_inode_lookup (fs, inum);
+    fprintf(hFile, "inode: %lu\n", (ULONG) inum);
+    fprintf(hFile, "%sAllocated\n",
+      (fs_inode->flags&FS_FLAG_ALLOC)?"":"Not ");
+
+    fprintf(hFile, "Group: %lu\n", (ULONG)ext2fs->grpnum);
+
+    if (fs_inode->link)
+        fprintf(hFile, "symbolic link to: %s\n", fs_inode->link);
+
+    fprintf(hFile, "uid / gid: %d / %d\n",
+      (int)fs_inode->uid, (int)fs_inode->gid);
+
+
+    make_ls(fs_inode->mode, ls, 12);
+    fprintf(hFile, "mode: %s\n", ls);
+
+	if (getu32(fs, ext2fs->dinode->i_flags)) {
+		fprintf(hFile, "Flags: ");
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_SECDEL)
+			fprintf(hFile, "Secure Delete, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_UNRM)
+			fprintf(hFile, "Undelete, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_COMP)
+			fprintf(hFile, "Compressed, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_SYNC)
+			fprintf(hFile, "Sync Updates, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_IMM)
+			fprintf(hFile, "Immutable, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_APPEND)
+			fprintf(hFile, "Append Only, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_NODUMP)
+			fprintf(hFile, "Do Not Dump, ");
+
+		if (getu32(fs, ext2fs->dinode->i_flags) & EXT2_IN_NOA)
+			fprintf(hFile, "No A-Time, ");
+		
+		fprintf(hFile, "\n");
+	}
+    fprintf(hFile, "size: %lu\n", (ULONG) fs_inode->size);
+    fprintf(hFile, "num of links: %lu\n", (ULONG) fs_inode->nlink);
+
+    fprintf(hFile, "\nInode Times:\n");
+	fprintf(hFile, "Accessed:\t%s", ctime(&fs_inode->atime));
+	fprintf(hFile, "File Modified:\t%s", ctime(&fs_inode->mtime));
+	fprintf(hFile, "Inode Modified:\t%s", ctime(&fs_inode->ctime));
+
+	if (fs_inode->dtime) 
+		fprintf(hFile, "Deleted:\t%s", ctime(&fs_inode->dtime));
+		
+    if (numblock > 0)
+        fs_inode->size = numblock * fs->file_bsize;
+
+    fprintf (hFile, "\nDirect Blocks:\n");
+
+    indir_idx = 0;
+    fs->file_walk(fs, fs_inode, 0, 0, flags, print_addr_act, (char *)hFile);
+
+    if (printidx != 0)
+        fprintf(hFile, "\n");
+
+    /* print indirect blocks */
+    if (indir_idx > 0) {
+        int i;
+        fprintf(hFile, "\nIndirect Blocks:\n");
+
+        printidx = 0;
+
+        for (i = 0; i < indir_idx; i++) {
+            fprintf(hFile, "%lu ", (unsigned long) indirl[i]);
+            if (++printidx == WIDTH) {
+                fprintf(hFile, "\n");
+                printidx = 0;
+            }
+        }
+        if (printidx != 0)
+            fprintf(hFile, "\n");
+    }
+
+    return;
+}
+
+
 /* ext2fs_close - close an ext2fs file system */
 
 static void 
@@ -985,7 +1131,7 @@ ext2fs_open(const char *name, unsigned char ftype)
     int     len;
 	FS_INFO *fs = &(ext2fs->fs_info);
 
-	if ((ftype & FSMASK) != EXT2FS_TYPE)   
+	if ((ftype & FSMASK) != EXT2FS_TYPE) 
 		error ("Invalid FS Type in ext2fs_open");
 
     /*
@@ -1015,7 +1161,7 @@ ext2fs_open(const char *name, unsigned char ftype)
 	 */
 
 	if (guessu16(fs, ext2fs->fs->s_magic, EXT2FS_FS_MAGIC))
-		error("%s is not EXT2FS", name);
+		error("Error: %s is not an EXT2FS file system", name);
 
     if (verbose) {
     	if (getu32(fs, ext2fs->fs->s_feature_ro_compat) & 
@@ -1074,6 +1220,7 @@ ext2fs_open(const char *name, unsigned char ftype)
 	fs->dent_walk = ext2fs_dent_walk;
 	fs->file_walk = ext2fs_file_walk;
 	fs->fsstat = ext2fs_fsstat;
+	fs->istat = ext2fs_istat;
     fs->close = ext2fs_close;
 
 	/* allocate buffers */
