@@ -264,7 +264,7 @@ is_lfn_name(fatfs_dentry_lfn *de)
 ** Walk the sectors of the partition. 
 **
 ** NOTE: This is by SECTORS and not CLUSTERS
-** _flags_ is only used for FS_FLAG_ALLOC / FS_FLAG_UNALLOC 
+** _flags_ is only used for FS_FLAG_ALLOC / FS_FLAG_UNALLOC / FS_FLAG_META
 **
 */
 void
@@ -292,6 +292,12 @@ fatfs_block_walk(FS_INFO *fs, DADDR_T start, DADDR_T last, int flags,
 		myflags = ((is_sectalloc(fatfs, addr)) ?
 			FS_FLAG_ALLOC : FS_FLAG_UNALLOC);
 
+		/* Anything less than the first data sector is either the FAT 
+		 * tables or super block stuff - therefore meta
+		 */
+		if (addr < fatfs->firstdatasect)
+			myflags |= FS_FLAG_META;
+
 		if ((flags & myflags) == myflags) {
 			fs_read_block(fs, fs_buf, fs->block_size, addr, "data block");
 			if (WALK_STOP == action(fs, addr, fs_buf->data, myflags, ptr)) {
@@ -392,7 +398,11 @@ set_secwest(void)
     tmptr = localtime(&t);
 
 #if HAVE_TIMEZONE
+#if defined (CYGWIN)
+    secwest = _timezone;
+#else
     secwest = timezone;
+#endif
 	/* account for daylight savings */
     if (tmptr->tm_isdst)
         secwest -= 3600;
@@ -428,6 +438,28 @@ dos2unixmode(u_int16_t attr)
 	return mode;
 }
 
+static void
+print_dosmode (FILE *hFile, int mode) 
+{
+    fprintf(hFile, "DOS Mode: ");
+
+	if (mode & FS_INODE_DIR) 
+		fprintf(hFile, "Directory");
+	else
+		fprintf(hFile, "File");
+
+
+	if ((mode & (MODE_IRUSR | MODE_IRGRP | MODE_IROTH)) == 0)
+		fprintf(hFile, ", Read Only");
+
+	if ((mode & (MODE_IWUSR | MODE_IWGRP | MODE_IWOTH)) == 0)
+		fprintf(hFile, ", Hidden");
+
+	fprintf(hFile, "\n");
+
+	return;
+}
+
 
 
 /*
@@ -473,6 +505,7 @@ fatfs_copy_inode(FATFS_INFO *fatfs, FS_INODE *fs_inode)
 
     fs_inode->dtime = 0;
 
+	fs_inode->seq = 0;
 
 	/* 
 	 * add the 8.3 file name 
@@ -486,6 +519,9 @@ fatfs_copy_inode(FATFS_INFO *fatfs, FS_INODE *fs_inode)
 	  i++) {
 		if ((i == 0) && (in->name[0] == FATFS_SLOT_DELETED))
 			fs_inode->name->name[0] = '_';
+		else if ((in->lowercase & FATFS_CASE_LOWER_BASE) && 
+		  (in->name[i] >= 'A') && (in->name[i] <= 'Z'))
+			fs_inode->name->name[i] = in->name[i] + 32;
 		else
 			fs_inode->name->name[i] = in->name[i];
 	}
@@ -495,7 +531,11 @@ fatfs_copy_inode(FATFS_INFO *fatfs, FS_INODE *fs_inode)
 		fs_inode->name->name[i++] = '.';
 		for (a = 0 ; (a < 3) && (in->ext[a] != 0) && (in->ext[a] != ' '); 
 		  a++, i++) {
-			fs_inode->name->name[i] = in->ext[a];
+			if ((in->lowercase & FATFS_CASE_LOWER_EXT) && 
+		  	  (in->ext[a] >= 'A') && (in->ext[a] <= 'Z'))
+				fs_inode->name->name[i] = in->ext[a] + 32;
+			else	
+				fs_inode->name->name[i] = in->ext[a];
 		}
 	}
 	fs_inode->name->name[i] = '\0';
@@ -539,6 +579,10 @@ fatfs_copy_inode(FATFS_INFO *fatfs, FS_INODE *fs_inode)
 
 	}
 
+	fs_inode->flags = ((in->name[0] == FATFS_SLOT_DELETED)?
+	  FS_FLAG_UNALLOC : FS_FLAG_ALLOC);
+
+	fs_inode->flags |= FS_FLAG_USED;
 
 	return;
 }
@@ -553,10 +597,7 @@ fatfs_make_root(FATFS_INFO *fatfs, FS_INODE *fs_inode)
 	int 		snum, cnum, i;
 	u_int32_t 	clust;
 
-	fs_inode->mode = (FS_INODE_DIR | 
-	  MODE_IXUSR | MODE_IXGRP | MODE_IXOTH |
-	  MODE_IRUSR | MODE_IRGRP | MODE_IROTH | 
-	  MODE_IWUSR | MODE_IWGRP | MODE_IWOTH);
+	fs_inode->mode = (FS_INODE_DIR);
 
 	fs_inode->nlink =  1;
 
@@ -884,11 +925,6 @@ fatfs_inode_lookup(FS_INFO *fs, INUM_T inum)
 		fatfs->dep = (fatfs_dentry *) &fatfs->dinodes->data[off]; 
 		if (fatfs_isdentry(fatfs, fatfs->dep)) {
 			fatfs_copy_inode(fatfs, fs_inode);
-
-			fs_inode->flags = ((fatfs->dep->name[0] == FATFS_SLOT_DELETED)?
-			  FS_FLAG_UNALLOC : FS_FLAG_ALLOC);
-
-			fs_inode->flags |= FS_FLAG_USED;
 		}
 		else {
 			error ("%d is not an inode", inum);
@@ -919,7 +955,10 @@ fatfs_file_walk(FS_INFO *fs, FS_INODE *fs_inode, u_int32_t type, u_int16_t id,
 	FS_BUF		*fs_buf;
 
 	fs_buf = fs_buf_alloc(fatfs->ssize);
-	size = fs_inode->size;
+	if (flags & FS_FLAG_SLACK)
+		size = roundup(fs_inode->size, fatfs->csize * fatfs->ssize);
+	else
+		size = fs_inode->size;
 
 	/* get the mask */
 	if (fs->ftype == MS16_FAT)
@@ -1099,6 +1138,72 @@ fatfs_fsstat(FS_INFO *fs, FILE *hFile)
 }
 
 
+/************************* istat *******************************/
+
+static int printidx = 0;
+#define WIDTH   8
+
+
+static u_int8_t
+print_addr_act (FS_INFO *fs, DADDR_T addr, char *buf,
+  int size, int flags, char *ptr)
+{
+    FILE *hFile = (FILE *)ptr;
+
+	fprintf(hFile, "%lu ", (unsigned long) addr);
+
+	if (++printidx == WIDTH) {
+		fprintf(hFile, "\n");
+		printidx = 0;
+    }
+
+    return WALK_CONT;
+}
+
+
+static void
+fatfs_istat (FS_INFO *fs, FILE *hFile, INUM_T inum, int numblock)
+{
+    FS_INODE *fs_inode;
+	FS_NAME *fs_name;
+    int flags = (FS_FLAG_AONLY | FS_FLAG_CONT | FS_FLAG_META |
+          FS_FLAG_ALLOC | FS_FLAG_UNALLOC | FS_FLAG_SLACK);
+
+    fs_inode = fatfs_inode_lookup (fs, inum);
+    fprintf(hFile, "Directory Entry: %lu\n", (ULONG) inum);
+    fprintf(hFile, "%sAllocated\n",
+      (fs_inode->flags&FS_FLAG_ALLOC)?"":"Not ");
+
+	print_dosmode(hFile, fs_inode->mode);
+	
+    fprintf(hFile, "size: %lu\n", (ULONG) fs_inode->size);
+	fprintf(hFile, "num of links: %lu\n", (ULONG) fs_inode->nlink);
+
+	if (fs_inode->name) {
+		fs_name = fs_inode->name;
+		fprintf(hFile, "Name: %s\n", fs_name->name);
+	}
+
+    fprintf(hFile, "\nDirectory Entry Times:\n");
+    fprintf(hFile, "Written:\t%s", ctime(&fs_inode->mtime));
+    fprintf(hFile, "Accessed:\t%s", ctime(&fs_inode->atime));
+    fprintf(hFile, "Created:\t%s", ctime(&fs_inode->ctime));
+
+    fprintf (hFile, "\nSectors:\n");
+
+    /* A bad hack to force a specified number of blocks */
+    if (numblock > 0)
+        fs_inode->size = numblock * fs->file_bsize;
+
+    fs->file_walk(fs, fs_inode, 0, 0, flags, print_addr_act, (char *)hFile);
+
+    if (printidx != 0)
+        fprintf(hFile, "\n");
+
+    return;
+}
+
+
 /* fatfs_close - close an fatfs file system */
 static void
 fatfs_close(FS_INFO *fs)
@@ -1142,7 +1247,7 @@ fatfs_open(const char *name, unsigned char ftype)
 
 	/* Check the magic value  and ID endian ordering */
 	if (guessu16(fs, fatsb->magic, FATFS_FS_MAGIC)) 
-			error ("%s is not FATFS", name);
+			error ("Error: %s is not a FATFS file system", name);
 
 	fatfs->ssize = getu16(fs, fatsb->ssize);
 	fatfs->csize = fatsb->csize;				/* cluster size */
@@ -1281,6 +1386,7 @@ fatfs_open(const char *name, unsigned char ftype)
     fs->dent_walk = fatfs_dent_walk;
     fs->file_walk = fatfs_file_walk;
 	fs->fsstat = fatfs_fsstat;
+	fs->istat = fatfs_istat;
     fs->close = fatfs_close;
 
 
