@@ -101,8 +101,8 @@ ffs_dinode_lookup(FFS_INFO *ffs, INUM_T inum)
 
     /*
      * Allocate/read the inode buffer on the fly.
-     */
-    if (ffs->dino_buf == 0)
+	 */
+	if (ffs->dino_buf == 0)
 		ffs->dino_buf = fs_buf_alloc(gets32(fs, ffs->fs->fs_bsize));
 
     addr = itod_lcl(fs, ffs->fs, inum);
@@ -118,6 +118,10 @@ ffs_dinode_lookup(FFS_INFO *ffs, INUM_T inum)
 
     memcpy((char *) ffs->dinode, ffs->dino_buf->data + offs, 
  	  sizeof(ffs_inode1));
+
+	ffs->inum = inum;
+
+	return;
 }
 
 /* ffs_dinode_free - destroy disk inode cache */
@@ -126,6 +130,8 @@ ffs_dinode_free(FFS_INFO *ffs)
 {
     if (ffs->dino_buf)
 		fs_buf_free(ffs->dino_buf);
+
+	ffs->inum = 0;
 }
 
 
@@ -138,8 +144,13 @@ ffs_copy_inode(FFS_INFO *ffs, FS_INODE *fs_inode)
 {
     int     i, j, count;
 	FS_INFO *fs = &(ffs->fs_info);
+	CGNUM_T cg_num;
+	ffs_cg *cg;
+	unsigned char *inosused = NULL;
+	INUM_T ibase;
 
 	fs_inode->flags = 0;
+	fs_inode->seq = 0;
 
 	/* If the symlink field is set from a previous run, then free it */
 	if (fs_inode->link) {
@@ -227,10 +238,13 @@ ffs_copy_inode(FFS_INFO *ffs, FS_INODE *fs_inode)
 					fs_read_block(fs, fs_buf, fs->block_size, 
 					  fs_inode->direct_addr[i], "link block");
 
-					strncpy (ptr, fs_buf->data, read_count);
+					memcpy (ptr, fs_buf->data, read_count);
 					count += read_count;
 					ptr = (char *)(int)ptr + read_count;
 				 }
+				 /* terminate the string */
+				 *ptr = '\0';
+
 				 fs_buf_free(fs_buf);
 			}
 		} /* end of symlink */
@@ -280,14 +294,36 @@ ffs_copy_inode(FFS_INFO *ffs, FS_INODE *fs_inode)
 				fs_read_block(fs, fs_buf, fs->block_size, 
 				  fs_inode->direct_addr[i], "link block");
 
-				strncpy (ptr, fs_buf->data, read_count);
+				memcpy (ptr, fs_buf->data, read_count);
 				count += read_count;
 				ptr = (char *)(int)ptr + read_count;
 			 }
+
+			/* terminate the string */
+			*ptr = '\0';
+
 			 fs_buf_free(fs_buf);
 		 }
 	}
 
+	/* set the flags */
+	cg_num = itog_lcl(fs, ffs->fs, ffs->inum);
+	cg = ffs_cgroup_lookup(ffs, cg_num);
+	inosused = (unsigned char *) cg_inosused_lcl(fs, cg);
+	ibase = cg_num * gets32(fs, ffs->fs->fs_ipg);
+
+	/* get the alloc flag */
+	fs_inode->flags = (isset(inosused, ffs->inum - ibase) ?
+	  FS_FLAG_ALLOC : FS_FLAG_UNALLOC);
+
+	/* link flag */
+	fs_inode->flags |= (fs_inode->nlink ?
+	  FS_FLAG_LINK : FS_FLAG_UNLINK);
+  
+	/* used/unused */
+	fs_inode->flags |= (fs_inode->ctime ? 
+	  FS_FLAG_USED : FS_FLAG_UNUSED);
+
 } /* end of ffs_copy_inode */
 
 
@@ -297,10 +333,6 @@ ffs_copy_inode(FFS_INFO *ffs, FS_INODE *fs_inode)
 static FS_INODE *
 ffs_inode_lookup(FS_INFO *fs, INUM_T inum)
 {
-	CGNUM_T cg_num;
-	ffs_cg *cg;
-	unsigned char *inosused = NULL;
-	INUM_T ibase;
     FFS_INFO *ffs = (FFS_INFO *) fs;
 	FS_INODE	*fs_inode;
 
@@ -311,25 +343,6 @@ ffs_inode_lookup(FS_INFO *fs, INUM_T inum)
     fs_inode = fs_inode_alloc(FFS_NDADDR, FFS_NIADDR);
     ffs_copy_inode(ffs, fs_inode);
 
-	/* We also need to figure out the flags associated with it */
-
-	cg_num = itog_lcl(fs, ffs->fs, inum);
-	cg = ffs_cgroup_lookup(ffs, cg_num);
-	inosused = (unsigned char *) cg_inosused_lcl(fs, cg);
-	ibase = cg_num * gets32(fs, ffs->fs->fs_ipg);
-
-	/* get the alloc flag */
-	fs_inode->flags = (isset(inosused, inum - ibase) ?
-	  FS_FLAG_ALLOC : FS_FLAG_UNALLOC);
-
-	/* link flag */
-	fs_inode->flags |= (fs_inode->nlink ?
-	  FS_FLAG_LINK : FS_FLAG_UNLINK);
-  
-	/* used/unused */
-	fs_inode->flags |= (fs_inode->ctime ? 
-	  FS_FLAG_USED : FS_FLAG_UNUSED);
-
     return (fs_inode);
 
 } /* end of ffs_inode_lookup */
@@ -809,6 +822,20 @@ ffs_fsstat(FS_INFO *fs, FILE *hFile)
           (((ULONG)cgbase_lcl(fs, sb, i + 1) - 1) < fs->last_block) ?
 		  ((ULONG)cgbase_lcl(fs, sb, i + 1) - 1) : fs->last_block);
 
+		/* The first group is special because the first 16 sectors are
+		 * reserved for the boot block.  
+		 * the next contains the primary Super Block 
+		 */
+		if (!i) {
+			fprintf(hFile, "    Boot Block: 0 - %lu\n",
+			  (ULONG)(15 * 512 / fs->block_size));
+
+			fprintf(hFile, "    Super Block: %lu - %lu\n",
+			  (ULONG)(16 * 512 / fs->block_size),
+			  (ULONG)(16 * 512 / fs->block_size) + 
+			  ((roundup(sizeof(ffs_sb), fs->block_size) / fs->block_size) - 1));
+		}
+
 		fprintf(hFile, "    Super Block: %lu - %lu\n",
 		  (ULONG)cgsblock_lcl(fs, sb, i),
 		  (ULONG)(cgsblock_lcl(fs, sb, i) + 
@@ -825,7 +852,17 @@ ffs_fsstat(FS_INFO *fs, FILE *hFile)
 		  ((roundup(gets32(fs, sb->fs_ipg)*sizeof(ffs_inode1), fs->block_size) 
 		  / fs->block_size) - 1)));
 
-		fprintf(hFile, "    Data Fragments: %lu - %lu\n",
+		fprintf(hFile, "    Data Fragments: ");
+
+		/* For all groups besides the first, the space before the
+		 * super block is also used for data
+		 */
+		if (i)
+			fprintf(hFile, "%lu - %lu, ",
+			  (ULONG)cgbase_lcl(fs, sb, i), 
+			  (ULONG)cgsblock_lcl(fs, sb, i) - 1);
+			  
+		fprintf(hFile, "%lu - %lu\n",
 		  (ULONG)cgdmin_lcl(fs, sb, i),
 		  (ULONG)(cgbase_lcl(fs, sb, i+1) - 1));
 
@@ -835,6 +872,122 @@ ffs_fsstat(FS_INFO *fs, FILE *hFile)
 }
 
 
+
+
+
+/************************* istat *******************************/
+
+static int printidx = 0;
+#define WIDTH   8
+
+/* indirect block accounting */
+#define INDIR_SIZ   64
+static DADDR_T indirl[INDIR_SIZ];
+static  unsigned char   indir_idx;
+
+
+static u_int8_t
+print_addr_act (FS_INFO *fs, DADDR_T addr, char *buf,
+  int size, int flags, char *ptr)
+{   
+	FILE *hFile = (FILE *)ptr;
+
+    if (flags & FS_FLAG_CONT) {
+        int i, s;
+        /* cycle through the fragments if they exist */
+        for (i = 0, s = size; s > 0; s-= fs->block_size, i++) {
+
+            /* sparse file */
+            if (addr)
+                fprintf(hFile, "%lu ", (unsigned long) addr + i);
+            else  
+                fprintf (hFile, "0 ");
+
+            if (++printidx == WIDTH) {
+                fprintf(hFile, "\n");
+                printidx = 0;
+            }
+        }
+    }
+
+    /* this must be an indirect block pointer, so put it in the list */
+    else if (flags & FS_FLAG_META) {
+        if (indir_idx < INDIR_SIZ)
+            indirl[indir_idx++] = addr;
+    }
+    return WALK_CONT;
+}
+
+
+static void
+ffs_istat (FS_INFO *fs, FILE *hFile, INUM_T inum, int numblock)
+{
+	FFS_INFO *ffs = (FFS_INFO *) fs;
+	FS_INODE *fs_inode;
+	char ls[12];
+    int flags = (FS_FLAG_AONLY | FS_FLAG_CONT | FS_FLAG_META |
+          FS_FLAG_ALLOC | FS_FLAG_UNALLOC );
+
+	fs_inode = ffs_inode_lookup (fs, inum);
+	fprintf(hFile, "inode: %lu\n", (ULONG) inum);
+	fprintf(hFile, "%sAllocated\n", 
+	  (fs_inode->flags&FS_FLAG_ALLOC)?"":"Not ");
+
+	fprintf(hFile, "Group: %lu\n", (ULONG)ffs->cg_num);
+
+    if (fs_inode->link)
+        fprintf(hFile, "symbolic link to: %s\n", fs_inode->link);
+
+    fprintf(hFile, "uid / gid: %d / %d\n", 
+	  (int)fs_inode->uid, (int)fs_inode->gid);
+
+
+    make_ls(fs_inode->mode, ls, 12);
+    fprintf(hFile, "mode: %s\n", ls);
+
+	fprintf(hFile, "size: %lu\n", (ULONG) fs_inode->size);
+	fprintf(hFile, "num of links: %lu\n", (ULONG) fs_inode->nlink);
+
+	fprintf(hFile, "\nInode Times:\n");
+	fprintf(hFile, "Accessed:\t%s", ctime(&fs_inode->atime));
+	fprintf(hFile, "File Modified:\t%s", ctime(&fs_inode->mtime));
+	fprintf(hFile, "Inode Modified:\t%s", ctime(&fs_inode->ctime));
+
+    /* A bad hack to force a specified number of blocks */
+    if (numblock > 0)
+        fs_inode->size = numblock * fs->file_bsize;
+
+	fprintf (hFile, "\nDirect Blocks:\n");
+
+	indir_idx = 0;
+    fs->file_walk(fs, fs_inode, 0, 0, flags, print_addr_act, (char *)hFile);
+
+    if (printidx != 0)
+        fprintf(hFile, "\n");
+
+    /* print indirect blocks */
+    if (indir_idx > 0) {
+        int i;
+        fprintf(hFile, "\nIndirect Blocks:\n");
+
+        printidx = 0;
+
+        for (i = 0; i < indir_idx; i++) {
+            fprintf(hFile, "%lu ", (unsigned long) indirl[i]);
+            if (++printidx == WIDTH) {
+                fprintf(hFile, "\n");
+                printidx = 0;
+            }
+        }
+        if (printidx != 0)
+            fprintf(hFile, "\n");
+    }
+
+	return;
+}
+
+
+
 /* ffs_close - close a fast file system */
 static void 
 ffs_close(FS_INFO *fs)
@@ -887,7 +1040,7 @@ ffs_open(const char *name, unsigned char ftype)
 
 	/* check the magic and figure out the endian ordering */
 	if (guessu32(fs, ffs->fs->fs_magic, FFS_FS_MAGIC))
-		error("%s is not FFS", name);
+		error("Error: %s is not a FFS file system", name);
 
     /*
      * Translate some filesystem-specific information to generic form.
@@ -922,6 +1075,7 @@ ffs_open(const char *name, unsigned char ftype)
     fs->dent_walk = ffs_dent_walk;
     fs->file_walk = ffs_file_walk;
 	fs->fsstat = ffs_fsstat;
+	fs->istat = ffs_istat;
     fs->close = ffs_close;
 
 	/* allocate the dinode buffer */
