@@ -5,13 +5,11 @@ extern FS_DATA *fs_data_alloc(u_int8_t);
 extern FS_DATA *fs_data_getnew_attr(FS_DATA *, u_int8_t);
 extern void fs_data_clear_list(FS_DATA *);
 
-extern void fs_data_size_str(FS_DATA *, int32_t);
-
 extern FS_DATA *fs_data_put_str(FS_DATA *, char *, u_int32_t, u_int16_t, 
   DADDR_T *, int);
 
-extern FS_DATA *fs_data_put_run(FS_DATA *, u_int64_t, char *, u_int32_t, 
-  u_int16_t, FS_DATA_RUN *, u_int64_t, u_int64_t);
+extern FS_DATA *fs_data_put_run(FS_DATA *, u_int64_t, u_int64_t,
+  FS_DATA_RUN *, char *, u_int32_t, u_int16_t, u_int64_t);
 
 extern FS_DATA *fs_data_lookup(FS_DATA *, u_int32_t, u_int16_t);
 
