@@ -258,13 +258,15 @@ struct FS_INFO {
     void    (*file_walk) (FS_INFO *, FS_INODE *, u_int32_t, u_int16_t, int, 
 	  FS_FILE_WALK_FN, char *);
 	void	(*fsstat) (FS_INFO *, FILE *);
+	void	(*istat) (FS_INFO *, FILE *, INUM_T, int);
     void    (*close) (FS_INFO *);
 };
 
 /* flag for FS_INFO flags */
-#define FS_HAVE_DTIME	0x01
-#define FS_LIT_ENDIAN	0x02
-#define FS_BIG_ENDIAN	0x04
+#define FS_LIT_ENDIAN	0x01
+#define FS_BIG_ENDIAN	0x02
+#define FS_HAVE_DTIME	0x04
+#define FS_HAVE_SEQ		0x08
 
 /* values for flags passed to the walk functions */
 #define FS_FLAG_LINK	(1<<0)		/* link count > 0 */
@@ -320,6 +322,7 @@ struct FS_DATA_RUN {
 	u_int8_t flags;
 };
 #define FS_DATA_FILLER	0x1
+#define FS_DATA_SPARSE	0x2
 
 struct FS_DATA {
 	FS_DATA		*next;
@@ -350,7 +353,8 @@ struct FS_DATA {
 struct FS_NAME {
 	FS_NAME *next;
 	char name[256];
-	u_int32_t parent;
+	u_int64_t par_inode;
+	u_int32_t par_seq;
 };
 
 struct FS_INODE {
@@ -368,6 +372,7 @@ struct FS_INODE {
     DADDR_T *indir_addr;		/* indirect blocks */
     int     indir_count;		/* number of blocks */
 
+	u_int32_t	seq;			/* sequence number (NTFS Only) */
 	FS_DATA	*attr;				/* additional attributes for NTFS */
 	FS_NAME *name;
 	char	*link;				/* used if this is a symbolic link */
@@ -388,7 +393,7 @@ struct FS_DENT {
     u_int16_t nextLink;  	/* distance until next entry
 							** according to the actual entry ptr */
     u_int16_t reclen;    	/* actual length of record based on name len */
-    unsigned char ent_type; /* dir, file etc FS_DENT_???*/
+    u_int8_t ent_type; 		/* dir, file etc FS_DENT_???*/
     u_int16_t namlen;   	/* length of actual name */
     ULONG maxnamlen;		/* size allocated to name*/
     char *name;     		/* name of file, directory etc
@@ -417,7 +422,9 @@ struct FS_DENT {
 /* ascii representation of above types */
 extern char fs_dent_str[FS_DENT_MAX_STR][2];
 
-
+/* string that is prepended to orphan FAT & NTFS files when the file
+ * name is known, but the parent is not */
+#define ORPHAN_STR "-ORPHAN_FILE-"
 
 /* Type of file in the mode field of Inodes.  FAT and NTFS are converted
  * to this mode value */
@@ -460,11 +467,14 @@ extern char fs_inode_str[FS_INODE_MAX_STR][2];
 extern FS_DENT *fs_dent_alloc(ULONG);
 extern FS_DENT *fs_dent_realloc(FS_DENT *, ULONG);
 extern void fs_dent_free(FS_DENT *);
-extern void fs_dent_print(FILE *, FS_DENT *, int, FS_INFO *, char *, char *);
-extern void fs_dent_print_long(FILE *, FS_DENT *, int, FS_INFO *, char *, char *);
-extern void fs_dent_print_mac(FILE *, FS_DENT *, int, FS_INFO *, char *);
+extern void fs_dent_print(FILE *, FS_DENT *, int, FS_INFO *, FS_DATA *);
+extern void fs_dent_print_long(FILE *, FS_DENT *, int, FS_INFO *, FS_DATA *);
+extern void fs_dent_print_mac(FILE *, FS_DENT *, int, FS_INFO *, 
+  FS_DATA *fs_data, char *);
 
 extern void make_ls(mode_t, char *, int);
+extern void fs_print_day (FILE *, time_t);
+extern void fs_print_time (FILE *, time_t);
 
 /*
 ** Is this string a "." or ".."
