@@ -27,7 +27,8 @@ FS_TYPES fs_open_table[] = {
 	{"fat32", MS32_FAT, "FAT32"},
     {"freebsd", FFS_1, "FreeBSD FFS"},
 	{"linux-ext2", EXT2FS, "Linux EXT2FS"},
-//	{"ntfs", NTFS, "NTFS"},
+	{"linux-ext3", EXT3FS_1, "Linux EXT3FS"},
+	{"ntfs", NTFS, "NTFS"},
     {"openbsd", FFS_1, "OpenBSD FFS"},
 	{"solaris", FFS_2, "Solaris FFS"},
 	{0},
